import subprocess

class FirewallEnforcer:
    def __init__(self, dry_run=True):
        """
        dry_run=True  → tidak benar-benar block (AMAN untuk riset)
        dry_run=False → aktifkan iptables (butuh sudo)
        """
        self.dry_run = dry_run

    def block_ip(self, ip):
        if self.dry_run:
            print(f"[ENFORCEMENT] (DRY RUN) Block IP: {ip}")
            return

        cmd = ["sudo", "iptables", "-A", "INPUT", "-s", ip, "-j", "DROP"]
        subprocess.run(cmd, check=False)
        print(f"[ENFORCEMENT] IP {ip} BLOCKED")

    def unblock_ip(self, ip):
        if self.dry_run:
            print(f"[ENFORCEMENT] (DRY RUN) Unblock IP: {ip}")
            return

        cmd = ["sudo", "iptables", "-D", "INPUT", "-s", ip, "-j", "DROP"]
        subprocess.run(cmd, check=False)
        print(f"[ENFORCEMENT] IP {ip} UNBLOCKED")



# palasik/core/plugin_loader.py

import importlib
import pkgutil
from pathlib import Path


class PluginLoader:
    """
    Loader plugin PALASIK.
    Mencari dan memuat plugin dari folder plugins/.
    """

    def __init__(self, plugins_path="plugins"):
        self.plugins_path = plugins_path

    def load(self):
        plugins = []

        if not Path(self.plugins_path).exists():
            return plugins

        for _, name, _ in pkgutil.iter_modules([self.plugins_path]):
            module_path = f"{self.plugins_path}.{name}.plugin"
            try:
                module = importlib.import_module(module_path)
                plugin_class = getattr(module, "Plugin", None)
                if plugin_class:
                    plugins.append(plugin_class())
            except Exception as e:
                # Plugin error tidak mematikan engine
                continue

        return plugins



# palasik/core/registry.py

class PluginRegistry:
    """
    Registry untuk menyimpan dan mengelola plugin.
    """

    def __init__(self):
        self._plugins = {}

    def register(self, plugin):
        name = plugin.name()
        if name in self._plugins:
            raise ValueError(f"Plugin '{name}' already registered")
        self._plugins[name] = plugin

    def get(self, name):
        return self._plugins.get(name)

    def all(self):
        return list(self._plugins.values())



# palasik/policy/base.py

from abc import ABC, abstractmethod

class PolicyEngine(ABC):
    """
    Kontrak policy engine PALASIK.
    """

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def decide(self, trust_score: float, event: dict, context) -> str:
        """
        Return decision:
        - 'ALLOW'
        - 'DENY'
        """
        pass



# palasik/trust/base.py

from abc import ABC, abstractmethod

class TrustEvaluator(ABC):
    """
    Kontrak evaluator kepercayaan device / event.
    """

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def evaluate(self, event: dict, context) -> float:
        """
        Return trust score (0.0 - 1.0)
        """
        pass



#INTI PALASIK - Menilai device berdasarkan perilaku, bukan identitas.
# palasik/core/trust_engine.py

class TrustEngine:
    def __init__(self):
        self.trust_map = {}

    def evaluate(self, device):
        ip = device["ip"]

        # Trust awal
        trust = self.trust_map.get(ip, 0.5)

        # Parameter evaluasi
        protocol = device.get("protocol", "UNKNOWN")
        freq = device.get("count", 1)

        # Penilaian berbasis protokol
        if protocol in ["MQTT", "HTTP"]:
            trust += 0.02
        else:
            trust -= 0.05

        # Penilaian berbasis frekuensi
        if freq > 10:
            trust -= 0.1

        # Clamp nilai trust
        trust = max(0.0, min(1.0, trust))

        # Simpan state
        self.trust_map[ip] = trust

        return {
            "ip": ip,
            "trust_score": round(trust, 2),
            "protocol": protocol
        }

#📌 PENTING SECARA AKADEMIK

#Tidak pakai ML dulu (Phase 3)



# palasik/core/engine.py

from palasik.core.registry import PluginRegistry
from palasik.core.decision import Decision


class PalasikEngine:
    """
    Engine utama PALASIK.
    Enforcement point Zero Trust.
    """

    def __init__(self, context):
        self.context = context
        self.registry = PluginRegistry()
        self.running = False

    def register_plugin(self, plugin):
        self.registry.register(plugin)

    def start(self):
        self.running = True
        for plugin in self.registry.all():
            plugin.on_start(self.context)

    def emit(self, event: dict):
        """
        Enforcement point utama.
        Hanya event ALLOW yang diteruskan.
        """

        trust_score = self.context.trust.evaluate(event, self.context)
        decision = self.context.policy.decide(trust_score, event, self.context)

        # 🔐 ENFORCEMENT
        if decision != Decision.ALLOW.value:
            self.context.logger.info(
                f"Event blocked by policy | trust={trust_score} | event={event}"
            )
            return

        # ✅ EVENT ALLOW → plugin
        for plugin in self.registry.all():
            plugin.on_event(event, self.context)

        # ✅ EVENT ALLOW → HTTP (opsional)
        http_adapter = getattr(self.context, "http_adapter", None)
        if http_adapter:
            http_adapter.forward(event)

    def stop(self):
        for plugin in self.registry.all():
            plugin.on_stop(self.context)
        self.running = False



# palasik/policy/allow_deny.py

from palasik.policy.base import PolicyEngine

class AllowDenyPolicy(PolicyEngine):

    def __init__(self, threshold: float = 0.5):
        self.threshold = threshold

    def name(self):
        return "allow_deny_policy"

    def decide(self, trust_score: float, event: dict, context) -> str:
        if trust_score >= self.threshold:
            return "ALLOW"
        return "DENY"



# palasik/enforcement/enforcer.py

def enforce(ip, action):
    if action == "ALLOW":
        return f"{ip} allowed"

    elif action == "MONITOR":
        return f"👀{ip} under monitoring"

    elif action == "RESTRICT":
        return f"⚠️ {ip} rate-limited (simulated)"

    elif action == "QUARANTINE":
        return f"🚫{ip} isolated (simulated)"

    return "UNKNOWN ACTION"



# palasik/core/plugin.py

from abc import ABC, abstractmethod

class PalasikPlugin(ABC):
    """
    Kontrak dasar untuk seluruh plugin PALASIK.
    """

    @abstractmethod
    def name(self) -> str:
        """
        Nama unik plugin.
        """
        pass

    @abstractmethod
    def version(self) -> str:
        """
        Versi plugin.
        """
        pass

    @abstractmethod
    def on_start(self, context):
        """
        Dipanggil saat agent mulai.
        """
        pass

    @abstractmethod
    def on_event(self, event: dict, context):
        """
        Dipanggil saat event diterima.
        """
        pass

    @abstractmethod
    def on_stop(self, context):
        """
        Dipanggil saat agent berhenti.
        """
        pass



# palasik/core/logger.py

class Logger:
    """
    Logger default PALASIK.
    Bisa digantikan plugin logger.
    """

    def info(self, message: str):
        print(f"[INFO] {message}")

    def warning(self, message: str):
        print(f"[WARN] {message}")

    def error(self, message: str):
        print(f"[ERROR] {message}")



# palasik/core/config.py

import os
import yaml


class ConfigLoader:
    def __init__(self, config_file: str | None = None):
        self.config = {}

        if config_file:
            self.load_yaml(config_file)

        self.load_env()

    def load_yaml(self, path: str):
        if not os.path.exists(path):
            raise FileNotFoundError(f"Config file not found: {path}")

        with open(path, "r") as f:
            data = yaml.safe_load(f) or {}
            self.config.update(data)

    def load_env(self):
        broker = self.config.setdefault("palasik", {}).setdefault("broker", {})
        policy = self.config.setdefault("palasik", {}).setdefault("policy", {})

        broker["host"] = os.getenv("PALASIK_BROKER_HOST", broker.get("host"))
        broker["port"] = int(os.getenv("PALASIK_BROKER_PORT", broker.get("port", 1883)))

        policy["threshold"] = float(
            os.getenv("PALASIK_POLICY_THRESHOLD", policy.get("threshold", 0.5))
        )

    def get(self, *keys, default=None):
        data = self.config
        for key in keys:
            data = data.get(key, {})
        return data or default



# palasik/core/agent.py

from palasik.core.context import PalasikContext
from palasik.core.engine import PalasikEngine
from palasik.core.config import ConfigLoader
from palasik.core.plugin_loader import PluginLoader


class PalasikAgent:
    """
    Entry point utama PALASIK.
    Dipakai oleh CLI dan runtime.
    """

    def __init__(self, plugins_path="plugins", config_file=None):
        # Load configuration
        self.config = ConfigLoader(config_file)

        # Build context
        self.context = PalasikContext(self.config)

        # Core engine
        self.engine = PalasikEngine(self.context)

        # Plugin loader
        self.loader = PluginLoader(plugins_path)

        # Setup optional adapters
        self._setup_optional_adapters()

    def _setup_optional_adapters(self):
        # HTTP adapter (optional)
        http_cfg = self.config.get("palasik", "http", default={})
        if http_cfg.get("enabled"):
            from palasik.adapters.http.adapter import HTTPAdapter
            self.context.http_adapter = HTTPAdapter(
                endpoint=http_cfg.get("endpoint"),
                timeout=http_cfg.get("timeout", 5),
            )

    def load_plugins(self):
        plugins = self.loader.load()
        for plugin in plugins:
            self.engine.register_plugin(plugin)

    def start(self):
        self.engine.start()

    def emit(self, event: dict):
        self.engine.emit(event)

    def stop(self):
        self.engine.stop()



# palasik/adapters/mqtt/adapter.py

import json
import paho.mqtt.client as mqtt

class MQTTAdapter:
    """
    MQTT Adapter untuk PALASIK.
    Mengubah pesan MQTT menjadi event PALASIK.
    """

    def __init__(
        self,
        agent,
        broker="localhost",
        port=1883,
        topic="palasik/sensor/#"
    ):
        self.agent = agent
        self.broker = broker
        self.port = port
        self.topic = topic

        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

    def on_connect(self, client, userdata, flags, rc):
        print("[MQTT] Connected with result code", rc)
        client.subscribe(self.topic)

    def on_message(self, client, userdata, msg):
        try:
            payload = msg.payload.decode()
            data = json.loads(payload)

            event = {
                "type": "mqtt",
                "topic": msg.topic,
                "value": data.get("value"),
                "raw": data
            }

            self.agent.emit_event(event)

        except Exception as e:
            print("[MQTT] Error processing message:", e)

    def start(self):
        print("[MQTT] Starting adapter...")
        self.client.connect(self.broker, self.port, 60)
        self.client.loop_start()

    def stop(self):
        print("[MQTT] Stopping adapter...")
        self.client.loop_stop()
        self.client.disconnect()



import requests

class HTTPAdapter:
    """
    Mengirim event ALLOW ke endpoint HTTP/Webhook.
    """

    def __init__(self, endpoint: str, headers=None, timeout: int = 5):
        self.endpoint = endpoint
        self.headers = headers or {"Content-Type": "application/json"}
        self.timeout = timeout

    def forward(self, event: dict) -> bool:
        try:
            resp = requests.post(
                self.endpoint,
                json=event,
                headers=self.headers,
                timeout=self.timeout,
            )
            resp.raise_for_status()
            return True
        except Exception:
            return False



import importlib
import subprocess
import sys

REQUIRED_PACKAGES = {
    "pandas": "pandas",
    "matplotlib": "matplotlib",
    "numpy": "numpy",
    "sklearn": "scikit-learn"
}

def ensure_dependencies():
    for module, package in REQUIRED_PACKAGES.items():
        try:
            importlib.import_module(module)
        except ImportError:
            print(f"📦 Installing missing dependency: {package}")
            subprocess.check_call([
                sys.executable,
                "-m",
                "pip",
                "install",
                package
            ])



#Untuk logging & eksperimen
# palasik/core/events.py
import csv
import os
from datetime import datetime

LOG_FILE = "data/logs/trust_log.csv"

def log_event(ip, trust, action):
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    print(f"[{timestamp[-8:]}] {ip} | trust={trust} | action={action}")

    file_exists = os.path.isfile(LOG_FILE)

    with open(LOG_FILE, mode="a", newline="") as f:
        writer = csv.writer(f)

        if not file_exists:
            writer.writerow(["timestamp", "ip", "trust_score", "action"])

        writer.writerow([timestamp, ip, trust, action])




# palasik/cli/main.py

import argparse
import sys
from pathlib import Path

from palasik.core.agent import PalasikAgent


def cmd_run(args):
    config_path = args.config

    if not Path(config_path).exists():
        print(f"[ERROR] Config file not found: {config_path}")
        sys.exit(1)

    print("[PALASIK] Starting agent...")
    agent = PalasikAgent(config_file=config_path)
    agent.load_plugins()
    agent.start()

    try:
        print("[PALASIK] Agent running. Press Ctrl+C to stop.")
        while True:
            pass
    except KeyboardInterrupt:
        print("\n[PALASIK] Stopping agent...")
        agent.stop()


def cmd_init(args):
    target = Path("config.yaml")
    if target.exists():
        print("[ERROR] config.yaml already exists")
        sys.exit(1)

    template = """palasik:
  broker:
    host: localhost
    port: 1883
    topic: palasik/sensor/#

  policy:
    threshold: 0.7

  http:
    enabled: false
    endpoint: "https://example.com/webhook"
    timeout: 5
"""

    target.write_text(template)
    print("[PALASIK] config.yaml created")


def main():
    parser = argparse.ArgumentParser(
        prog="palasik",
        description="PALASIK IoT Zero Trust Gateway",
    )

    subparsers = parser.add_subparsers(dest="command")

    # palasik run
    run_parser = subparsers.add_parser("run", help="Run PALASIK agent")
    run_parser.add_argument(
        "--config",
        default="config.yaml",
        help="Path to config.yaml",
    )
    run_parser.set_defaults(func=cmd_run)

    # palasik init
    init_parser = subparsers.add_parser("init", help="Create config.yaml template")
    init_parser.set_defaults(func=cmd_init)

    args = parser.parse_args()

    if not hasattr(args, "func"):
        parser.print_help()
        sys.exit(1)

    args.func(args)


if __name__ == "__main__":
    main()



# palasik/trust/simple.py

from palasik.trust.base import TrustEvaluator

class SimpleTrustEvaluator(TrustEvaluator):
    def name(self):
        return "simple_trust"

    def evaluate(self, event: dict, context) -> float:
        """
        Contoh sederhana:
        - sensor value <= 100 → trusted
        - sisanya → low trust
        """
        value = event.get("value", 0)

        if value <= 100:
            return 0.9
        return 0.2



# plugins/logger/plugin.py

from palasik.core.plugin import PalasikPlugin
from palasik.trust.simple import SimpleTrustEvaluator

class LoggerPlugin(PalasikPlugin):

    def __init__(self):
        self.trust = SimpleTrustEvaluator()

    def name(self):
        return "logger"

    def version(self):
        return "1.0.0"

    def on_start(self, context):
        print("[Logger] PALASIK Agent started")

    def on_event(self, event, context):
        trust_score = self.trust.evaluate(event, context)
        print(f"[Logger] Event={event} | Trust={trust_score}")

    def on_stop(self, context):
        print("[Logger] PALASIK Agent stopped")



"""
PALASIK Policy Engine
Otak pengambilan keputusan berbasis trust score
"""

class PolicyEngine:
    def decide(self, trust_score):
        if trust_score >= 0.8:
            return "ALLOW"
        elif trust_score >= 0.5:
            return "MONITOR"
        elif trust_score >= 0.3:
            return "RESTRICT"
        else:
            return "QUARANTINE"




import subprocess
import sys
import importlib

REQUIRED_PACKAGES = [
    "pandas",
    "matplotlib",
    "numpy"
]

def ensure_dependencies():
    """
    Pastikan semua dependency tersedia.
    Jika tidak ada → install via pip (dalam venv).
    """
    for pkg in REQUIRED_PACKAGES:
        try:
            importlib.import_module(pkg)
        except ImportError:
            print(f"[DEPENDENCY] Installing missing package: {pkg}")
            subprocess.check_call([
                sys.executable, "-m", "pip", "install", pkg
            ])



# palasik/core/loader.py

import importlib.util
import os
import inspect

from palasik.core.plugin import PalasikPlugin

class PluginLoader:
    """
    Auto discovery dan loader plugin PALASIK.
    """

    def __init__(self, plugins_path="plugins"):
        self.plugins_path = plugins_path

    def discover(self):
        plugins = []

        if not os.path.isdir(self.plugins_path):
            return plugins

        for name in os.listdir(self.plugins_path):
            plugin_dir = os.path.join(self.plugins_path, name)
            plugin_file = os.path.join(plugin_dir, "plugin.py")

            if not os.path.isfile(plugin_file):
                continue

            spec = importlib.util.spec_from_file_location(
                f"plugins.{name}",
                plugin_file
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            for _, obj in inspect.getmembers(module, inspect.isclass):
                if issubclass(obj, PalasikPlugin) and obj is not PalasikPlugin:
                    plugins.append(obj())

        return plugins



from palasik.policy.allow_deny import AllowDenyPolicy
from palasik.core.engine import PalasikEngine
from palasik.core.context import PalasikContext

class DummyTrust:
    def evaluate(self, event, context):
        return event["trust"]

class DummyPlugin:
    def __init__(self):
        self.called = False

    def on_event(self, event, context):
        self.called = True

def test_event_denied_not_reach_plugin():
    context = PalasikContext()
    context.trust = DummyTrust()
    context.policy = AllowDenyPolicy(threshold=0.7)

    engine = PalasikEngine(context)
    plugin = DummyPlugin()
    engine.register_plugin(plugin)

    engine.emit({"trust": 0.3})

    assert plugin.called is False



from palasik.policy.allow_deny import AllowDenyPolicy
from palasik.trust.simple import SimpleTrustEvaluator
from palasik.core.logger import Logger


class PalasikContext:
    """
    Context global PALASIK.
    Menyimpan state bersama: config, trust, policy, logger, adapter.
    """

    def __init__(self, config=None):
        self.config = config

        # Logger default
        self.logger = Logger()

        # ✅ Trust engine default (IMPLEMENTASI NYATA)
        self.trust = SimpleTrustEvaluator()

        # Policy engine default (configurable)
        threshold = 0.5
        if config:
            threshold = config.get("palasik", "policy", "threshold", default=0.5)

        self.policy = AllowDenyPolicy(threshold=threshold)

        # Adapter opsional
        self.http_adapter = None



#SIMULASI BERJALAN
# main.py
from palasik.core.dependency import ensure_dependencies
ensure_dependencies()

from palasik.core.device_registry import DeviceRegistry
from palasik.core.trust_engine import TrustEngine
from palasik.core.policy_engine import PolicyEngine
from palasik.core.events import log_event
from palasik.enforcement.enforcer import enforce
from palasik.enforcement.firewall import FirewallEnforcer
import time

registry = DeviceRegistry()
trust_engine = TrustEngine()
policy_engine = PolicyEngine()
enforcer = FirewallEnforcer(dry_run=True)

# Simulasi traffic
traffic = [
    ("192.168.1.10", "B8:27:EB:AA", "MQTT"),
    ("192.168.1.10", "B8:27:EB:AA", "MQTT"),
    ("192.168.1.20", "DC:A6:32:BB", "HTTP"),
    ("192.168.1.10", "B8:27:EB:ZZ", "UNKNOWN"),  # spoof
    ("192.168.1.10", "B8:27:EB:ZZ", "UNKNOWN"),
]

for ip, mac, proto in traffic:
    device = registry.register_or_update(ip, mac, proto)
    trust = trust_engine.evaluate(device)
    action = policy_engine.decide(trust["trust_score"])

    if action == "QUARANTINE":
        enforcer.block_ip(ip)

    result = enforce(ip, action)
    log_event(ip, trust["trust_score"], action)
    print("ENFORCE:", result)
    time.sleep(1)




# device_registry.py
# palasik/core/device_registry.py
import time

class DeviceRegistry:
    def __init__(self):
        self.devices = {}

    def register_or_update(self, ip, mac, protocol):
        now = time.time()

        if ip not in self.devices:
            self.devices[ip] = {
                "ip": ip,
                "mac": mac,
                "protocol": protocol,
                "first_seen": now,
                "last_seen": now,
                "count": 1
            }
        else:
            device = self.devices[ip]
            device["last_seen"] = now
            device["count"] += 1

            # update protocol jika berubah
            if protocol:
                device["protocol"] = protocol

        return self.devices[ip]

#phase1
#Tugas:1.Menyimpan & mengelola profil device BUKAN menilai trust
#📌 Catatan desain: Registry tidak pintar , Registry tidak membuat keputusan,Registry hanya menyimpan fakta
